
'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '38.5.3',
    'branch_nick': 'release-38-LTS',
    'revision_id': '0613838eea50056ad38456d06a836af09c5b208e',
    'revno': 14134
}
